Require Import Buildmathcomp.
Require  Ppform.
Require Lia.

Ltac ppsimplcc :=
  Ppform.ppsimpl_with Buildmathcomp.make_inj Buildmathcomp.simplifier.

Require Import ZArith.

Ltac pos_cst T :=
  match T with
  | xH => constr:(true)
  | xO ?X => pos_cst X 
  | xI ?X  => pos_cst X
  | _   => constr:(false)
  end.

Inductive Zcst :=
| Full 
| Partial (l : list Z)
.

Definition Zcst_of_bool (z :Z) (b : bool) :=
  if b then Full  else Partial nil.


Ltac list_app T l1 l2 :=
  match l1 with
  | nil => l2
  | cons ?e ?l => let r := list_app T l l2 in
                   constr:(@cons T e r)
  end.



Ltac Zcst_of_bin z1 z2 c1 c2 :=
  match constr:((c1, c2)) with
  | (Full , Full) => constr:(Full)
  | (Full , Partial ?l)  => constr:(Partial (z1::l))
  | (Partial ?l , Full)  => constr:(Partial (z2::l))
  | (Partial ?l1 , Partial ?l2) =>
    let l := list_app Z l1 l2 in
    constr:(Partial l)
  end.


Ltac list_of_Zcst z c acc :=
  match c with
  | Full => constr:(cons z acc)
  | Partial ?l => list_app Z l acc
  end.


Ltac zcst T :=
    match T with
    | Z0 => constr:(Full)
    | Zpos ?p =>
      let b := pos_cst p in
      eval lazy in (Zcst_of_bool T b)
    | Zneg ?p =>
      let b := pos_cst p in
      eval lazy in (Zcst_of_bool T b)
    | Z.succ ?z => zcst z
    | Z.add ?z1 ?z2 =>
      let c1 := zcst z1 in
      let c2 := zcst z2 in
      Zcst_of_bin z1 z2 c1 c2
    | Z.mul ?z1 ?z2 =>
      let c1 := zcst z1 in
      let c2 := zcst z2 in
      Zcst_of_bin z1 z2 c1 c2
    | Z.sub ?z1 ?z2 =>
      let c1 := zcst z1 in
      let c2 := zcst z2 in
      Zcst_of_bin z1 z2 c1 c2
    | Z.pow ?z1 ?z2 => 
      let c1 := zcst z1 in
      let c2 := zcst z2 in
      Zcst_of_bin z1 z2 c1 c2
    | Z.max ?z1 ?z2 => 
      let c1 := zcst z1 in
      let c2 := zcst z2 in
      Zcst_of_bin z1 z2 c1 c2
    |  _  => constr:(Partial nil)
    end.

Ltac collect_cst T :=
  let rec all_cst :=
      fun L T =>
        match type of T with
        | Z => let c := zcst T in
               list_of_Zcst T c L
        | Prop => 
          match T with
          | ?A /\ ?B => all_cst2 L A B
          | ?A -> ?B => all_cst2 L A B
          | ?A \/ ?B => all_cst2 L A B
          | Z.le ?A ?B => all_cst2 L A B
          | @eq Z ?A ?B => all_cst2 L A B
          | iff ?A ?B   => all_cst2 L A B
          |   _         => L
          end
        |  _  => L
        end
        with  all_cst2  :=
        fun L A B => 
          let l1 := all_cst L A in
          all_cst l1 B
          in
            all_cst (@nil Z) T.

      Ltac insert F e s :=
        match s with
        | nil => constr:(cons e s)
        | cons ?e' ?s' =>
          let cmp := eval compute in (F e e') in
              match cmp with
              | Lt => constr:(cons e s)
              | Eq => constr:(cons e s)
              | Gt => let s2 := insert F e s' in
                      constr:(cons e' s2)
              end
        end.
                      
      Definition cmpZ (z1 z2 : Z) : comparison :=
        Z.compare z2 z1.
        
      Ltac sort F L :=
        match L with
        | nil => L
        | cons ?e ?l =>
          let s := sort F l in
          insert F e s
        end.
          
      Ltac eval_all L :=
        match L with
        | nil => idtac
        | @cons _ ?x ?l =>
          let x':= eval compute in x in
              change x with x' ; eval_all l
        end.
      
      Ltac eval_all_cst :=
        match goal with
        | |- ?g => let l := collect_cst g in
                   let l' := sort cmpZ l in
                   eval_all l'
        end.

      Ltac ppsimpl := Ppform.ppsimpl ; eval_all_cst.

